/*	$Id: inode.c,v 1.53 2007-08-07 00:24:16 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/

//
// inode.c : "inodes" and "filehandles".
// 

#include <dllist.h>
#include <coda_types.h>
#include <coda.h>
#include <coda_nt.h>
#include <cnode.h>
#include <ctype.h>

// Prototypes for local routines

static void inode_cache_delete (coda_inode *inode);

// Global variable for purgeuser and flush

ULONG glo_vattr_cntr;

// A constant to help decide if we want to free or reuse and
// inode.   Number is a guess!

#define MIN_INODES 250 

#define CACHE_SIZE 1023
#define MAX_NHASH_SHIFT 3
struct dllist_head inode_fid_cache[CACHE_SIZE+1];
struct dllist_head inode_name_cache[CACHE_SIZE+1];

static ERESOURCE cache_lock;

// list of inode not pinned by wrapper -- inserted at tail.
// reused at head.  Nodes on this list may still be in the
// cache!  "released inodes" (e.g. deleted files) are 
// inserted at head.  (if the list is long enough, released
// inodes are freed instead of being put on the free list.)

struct dllist_head inode_freelist;
static int free_count; 

// list of all inodes ... so we can free all at umount time.
struct dllist_head inode_all;

// inode globals
static int inode_count;

// file handle globals
static coda_filehandle *fh_free;
static int fh_unique;

void inode_init()
{
    int i;
    
    // The cache ...
    for ( i = 0 ; i <= CACHE_SIZE ; i++ ) { 
	list_init(&(inode_fid_cache[i]));
	list_init(&(inode_name_cache[i]));
    }

    list_init(&inode_all);
    inode_count = 0;

    list_init(&inode_freelist);
    free_count = 0;

    // File handles ...
    fh_free = NULL;
    fh_unique = 1;

    // Global vattr number
    glo_vattr_cntr = 1;

    // Initialize the global resource lock
    ExInitializeResourceLite (&CodaWrapperLock);
    ExInitializeResourceLite (&cache_lock);
}

void inode_cleanup()
{
    ExDeleteResourceLite (&CodaWrapperLock);
    ExDeleteResourceLite (&cache_lock);
}

// Hash functions.

static int inode_fhash(CodaFid *fid)
{
    return (CACHE_SIZE & (coda_f2i(fid)));
}

static int inode_nhash(char *name, CodaFid *pfid) 
{
    int len = strlen(name);
    int val = inode_fhash(pfid) + len;
    int i;
    
    for (i=0; i<len; i++)
	val += tolower(name[i]) << (i & MAX_NHASH_SHIFT);

    return (CACHE_SIZE & val);
}

// Insert -- don't put on freelist if pinned or already there!

static void inode_cache_insert(coda_inode *inode)
{
    int fhash = inode_fhash(&inode->c_fid);
    int nhash = inode_nhash(inode->c_name, &(inode->c_parent->c_fid));

    CDEBUG ((D_INODE,
	     "inode_cache_insert: inode %s [%s] fhash = %d, nhash = %d\n",
	     inode->c_name, coda_f2s(&inode->c_fid), fhash, nhash));

    list_add(&inode->c_fidlist,  &(inode_fid_cache[fhash]));
    list_add(&inode->c_namelist, &(inode_name_cache[nhash]));
    if ( list_empty(&inode->c_freelist) && !inode_hasflag(inode, I_PINNED)) {
	list_add(&inode->c_freelist, inode_freelist.prev);
	free_count++;
    }
}

// Need to delete an inode from the cache.

static void inode_cache_delete (coda_inode *inode)
{
    LockCache();
    if ( !list_empty(&inode->c_fidlist) ) {
	list_del(&inode->c_fidlist);
    }
    if ( !list_empty(&inode->c_namelist) ) {
	list_del(&inode->c_namelist);
    }
    UnLockCache();
}


void inode_rename(coda_inode *inode, coda_inode *parent, char *name)
{

    // Remove from cache lists
    inode_cache_delete (inode);   // Calls LockCache() inside.

    LockCache();
    if (parent != inode->c_parent) {
	// Change counts
	list_del (&inode->c_childlist);
	inode->c_parent->c_child_cnt--;
#if DBG
	if (inode->c_parent->c_child_cnt < 0)
	    DbgPrint ("inode_rename: c_child_cnt %d, %s/%s\n",
		      inode->c_parent->c_child_cnt,
		      inode->c_parent->c_name, inode->c_name);
#endif
	inode->c_parent = parent;
	list_add (&inode->c_childlist, &inode->c_parent->c_children);
	parent->c_child_cnt++;
    }
    if (name != inode->c_name)
	strncpy (inode->c_name, name, CODA_MAXNAMLEN);

    // Reinsert
    inode_cache_insert (inode);
    UnLockCache();
}

// Utility routines

int inode_fideq(CodaFid *f1, CodaFid *f2)
{   
    return ((f1->opaque[0] == f2->opaque[0]) &&
            (f1->opaque[1] == f2->opaque[1]) &&
            (f1->opaque[2] == f2->opaque[2]) &&
            (f1->opaque[3] == f2->opaque[3]));
}

coda_inode *inode_find_fid(CodaFid *fid)
{
    int hash = inode_fhash(fid);
    struct dllist_head *lh, *le;
    coda_inode *inode;
    int max = inode_count;

    CDEBUG((D_INODE, "inode_find_fid: hash for %s is %d, max is %d\n", 
	    coda_f2s(fid), hash, max));

    LockCache();
    le = &(inode_fid_cache[hash]);
    lh = le;
    for ( le = lh->next ; (le != lh) && max ; le = le->next, max-- ) {
	inode = list_entry(le, coda_inode, c_fidlist);
	CDEBUG ((D_INODE | D_DETAILS,
		 "inode_find_fid: at %s\n", coda_f2s(&inode->c_fid)));
	if ( inode_fideq(&inode->c_fid, fid) ) {
	    UnLockCache();
	    return inode ; 
	}
    }
    UnLockCache();
    
    CDEBUG ((D_INODE, "inode_find_fid: not found, max= %d\n", max));
    return NULL;
}

coda_inode *inode_find_name(coda_inode *parent,
			    char *name,
			    int CaseInsensitive)
{
    int hash = inode_nhash(name, &parent->c_fid);
    struct dllist_head *lh, *le;
    coda_inode *inode;
    int max = inode_count;

    CDEBUG ((D_INODE, "inode_find_name: name= %s, parent= 0x%x,  nhash= %d\n",
	     name, parent, hash));

    LockCache();
    le = &(inode_name_cache[hash]);
    lh = le;
    for ( le = lh->next ; (le != lh) && max ; le = le->next, max-- ) {
	inode = list_entry(le, coda_inode, c_namelist);
	CDEBUG ((D_INODE | D_DETAILS,
		 "inode_find_name: at %s 0x%x (str %d, case %d)\n",
		 inode->c_name, inode->c_parent, strcmp(name, inode->c_name),
		 strcasecmp(name, inode->c_name)));
	if ( inode->c_parent != parent )
	    continue ;
	if ( (strcmp(name, inode->c_name) == 0)
	     || (CaseInsensitive && (strcasecmp(name, inode->c_name) == 0)) ) {
	    UnLockCache();
	    return inode;
	}
    }
    UnLockCache();

    CDEBUG ((D_INODE, "inode_find_name: not found, max=%d\n", max));
    return NULL;
}

// Update Fid if needed
void inode_update(coda_inode *inode, CodaFid *fid)
{
    int fhash;
    CDEBUG ((D_INODE, "inode_update: %s\n", inode->c_name));

    if (!inode_fideq(&inode->c_fid, fid)) {
	CDEBUG ((D_INODE, "inode_update: new fid %s\n", coda_f2s(fid)));
	inode->c_fid = *fid;
	fhash = inode_fhash(fid);
	LockCache();
	if (!list_empty(&inode->c_fidlist)) {
	    list_del(&inode->c_fidlist);
	} else {
	    DbgPrint("inode_update: inode not in fid hash, %s\n",
		     inode->c_name);
	}
	list_add(&inode->c_fidlist, &(inode_fid_cache[fhash]));
	UnLockCache();
    }
}



// Get a new inode!  Leaves it on the free list! (at the end)

coda_inode *inode_new(CodaFid *fid, coda_inode *parent, char *name,
		      int AddToCache)
{
    int fidhash, namehash;
    coda_inode *inode = NULL;

    CDEBUG ((D_INODE, "inode_new: %s\n", name));

    if ( !list_empty (&inode_freelist) && inode_count > MIN_INODES) {

	// Reuse an inode
	inode = list_entry(inode_freelist.next, coda_inode, c_freelist);
	
	if (inode->c_child_cnt) {
	    int num = free_count;
	    coda_inode *temp;

	    while ( num && inode->c_child_cnt ) {
		CDEBUG((D_INODE, "inode_new: not reusing 0x%x, has children\n",
			inode));
		temp = list_entry (inode->c_freelist.next, coda_inode,
				   c_freelist);
		inode = temp;
		num--;
	    }
	    if (!num || inode->c_child_cnt)
		inode = NULL;
	}
    }

    if (inode) {

	CDEBUG ((D_INODE, "inode_new: Reusing inode 0x%x\n", inode));

	inode_cache_delete (inode);  // Calls LockCache() inside.

	LockCache();
	list_del(&inode->c_freelist);
	free_count--;

	if ((inode->c_parent != inode) && (inode->c_parent->c_child_cnt > 0)) {
	    inode->c_parent->c_child_cnt--;
#if DBG
	    if (inode->c_parent->c_child_cnt < 0)
		DbgPrint ("inode_new: c_child_cnt %d, %s/%s\n",
			  inode->c_parent->c_child_cnt,
			  inode->c_parent->c_name, inode->c_name);
#endif
	    list_del(&inode->c_childlist);
	}
	UnLockCache();

    } else {

	// Create a new one.
    
	inode = (coda_inode *) coda_malloc(sizeof(*inode));
	if (!inode) {
	    DbgPrint("inode_new: Out of memory!\n");
	    return NULL;
	}

	LockCache();
	inode_count++;

	// Initialize fields that won't be zeroed later

	inode->c_magic = CODA_MAGIC;
	list_add(&inode->c_alllist, &inode_all);
	UnLockCache();
	ExInitializeResourceLite(&inode->c_lock);

	CDEBUG ((D_INODE, "inode_new: Made new inode 0x%x\n", inode));
    }

    // Make sure all lists are initialized (possibly again)
    list_init(&inode->c_fidlist);
    list_init(&inode->c_namelist);
    list_init(&inode->c_freelist);
    list_init(&inode->c_childlist);
    
    // Initialize all to zero
    RtlZeroMemory ((void *)&(inode->c_fh_list),
		   sizeof(*inode) -
		   ((int)(&inode->c_fh_list) - (int)(inode)));
		   

    // Root node points to itself ... others point to real parent
    if ( !parent ) { 
	inode->c_parent = inode;
    } else {
	inode->c_parent = parent;
	parent->c_child_cnt++;
	list_add(&inode->c_childlist, &inode->c_parent->c_children);
    }

    // Fill in non-zero elements 
    inode->c_fh_list = NULL;
    inode->c_fh_cnt = 0;
    inode->c_fid = *fid;
    list_init(&(inode->c_children));
    inode->c_child_cnt = 0;
    inode->c_ntvol = &coda_nt_volume;
    strncpy(inode->c_name, name, CODA_MAXNAMLEN-1);
    inode->c_name[CODA_MAXNAMLEN-1] = 0;

    // Initialize the event and lock
    KeInitializeEvent (&inode->c_event, NotificationEvent, FALSE);

    // Add at the end of the freelist and put in the cache if requested
    LockCache();
    list_add(&inode->c_freelist, inode_freelist.prev);
    free_count++;
    if (AddToCache)
	inode_cache_insert(inode);
    UnLockCache();
    
    return inode;
}

// We are really done with this inode.  Called when an
// inode needs to be not found again ... e.g. when deleted ...

void inode_release(coda_inode *inode)
{
    CDEBUG ((D_INODE, "inode_release: Releasing %s\n", inode->c_name));

#ifdef _BIG_FILE_
    // Clean up the big file struct
    if (inode->c_big_filep) {
      coda_free(inode->c_big_filep);
      inode->c_big_filep = NULL;
    }
#endif

    if (inode->c_child_cnt) {
	// Make sure we don't leave dangling pointers
	DbgPrint ("inode_release: "
		  "inode non-zero c_child_cnt field (%d, %s)\n",
		  inode->c_child_cnt, inode->c_name);
	inode->c_status = I_NEEDREL;
	return;
    }

    if (inode->c_lnk_file != NULL) {
	coda_free(inode->c_lnk_file);
	inode->c_lnk_file = NULL;
	inode->c_lnk_state = SLNK_NONE;
    }

    inode->c_status = I_REL;
    if (inode->c_parent != inode) {
	LockCache();
	list_del(&inode->c_childlist);
	inode->c_parent->c_child_cnt--;
#if DBG
	if (inode->c_parent->c_child_cnt < 0)
	    DbgPrint ("inode_release: c_child_cnt %d, %s/%s\n",
		      inode->c_parent->c_child_cnt,
		      inode->c_parent->c_name, inode->c_name);
#endif
	UnLockCache();
    }

    // Clean up parent if waiting for this inode to be released
    if ((inode->c_parent != inode) && (inode->c_parent->c_child_cnt == 0)
	&& (inode->c_parent->c_status == I_NEEDREL))
	inode_release(inode->c_parent);

    if (free_count > MIN_INODES) {
	inode_free(inode);
	return;
    }

    // Add node to the front of the free list

    inode_cache_delete (inode);   // Calls LockCache() inside.
    inode->c_name[0] = 0;  
    inode->c_parent = inode;
    LockCache();
    if ( !list_empty(&inode->c_freelist) ) {
	list_del(&inode->c_freelist);
	free_count--;
    }

    list_add(&inode->c_freelist, &inode_freelist);
    free_count++;

    UnLockCache();
}

// Inode pin and unpin ... keep pinned inodes off the free list!

void inode_pin(coda_inode *inode)
{
    CDEBUG ((D_INODE, "inode_pin:\n"));
    LockCache();
    if ( !list_empty(&inode->c_freelist) ) {
	list_del(&inode->c_freelist);
	free_count--;
    } 
    
    inode_setflag (inode, I_PINNED);
    coda_nt_volume.v_refcount++;
    UnLockCache();
}

void inode_unpin(coda_inode *inode)
{
    int rc;

    CDEBUG ((D_INODE, "inode_unpin:\n"));

    if (inode_hasflag(inode, I_DELETE)) {
	CDEBUG ((D_WRAPPER, "inode_unpin: Closing file.\n"));
	if (inode->c_open_cnt > 0) {
	    CDEBUG((D_FILE, "inode_unpin: Closing inode file object.\n")); 
	    if (inode->c_object != NULL) {
		// Close the object!
		inode_close_handle(inode);
	    } else {
		DbgPrint("inode_unpin: Closing NULL c_object (%s), bad news\n",
			 inode->c_name);
	    }
	}
	rc = venus_close (&(inode->c_fid), C_O_WRITE, GetNtUID());
    }
    
    LockCache();
    list_add(&inode->c_freelist, inode_freelist.prev);
    free_count++;
    inode_unsetflag (inode, I_PINNED);
    coda_nt_volume.v_refcount--;
    UnLockCache();

    if (inode_hasflag(inode, I_PURGE)) {
	CDEBUG ((D_INODE, "inode_unpin: releaseing inode\n"));
	inode_release(inode);
    }
}

// Actually delete the inode space!

void inode_free(coda_inode *inode)
{
    CDEBUG ((D_INODE, "inode_free:\n"));
    inode_cache_delete (inode);   // Calls LockCache() inside.
    LockCache();
    if ( !list_empty(&inode->c_freelist) ) {
	list_del(&inode->c_freelist);
	free_count--;
    }
    if ( !list_empty(&inode->c_alllist) )
	list_del(&inode->c_alllist);
    inode_count --;
    UnLockCache();
    
    ExDeleteResourceLite (&inode->c_lock);
    coda_free(inode);

}

// Free all the inodes ...

void inode_freeall()
{
    struct dllist_chain *iptr;
    coda_inode *inode;
    int cnt;

    CDEBUG ((D_INODE, "inode_freeall:\n"));

    iptr = inode_all.next;
    cnt = inode_count;
    while (cnt-- > 0) {
	inode = list_entry(iptr, coda_inode, c_alllist);
	iptr = iptr->next;
	CDEBUG ((D_INODE, " freeall: %s\n", coda_f2s(&inode->c_fid)));
	inode_free (inode);
    }
}

void wi_close_handle ( PDEVICE_OBJECT pdev, PVOID Arg )
{
    coda_inode *inode = Arg;

    CDEBUG ((D_INODE, "wi_close_handle: inode->c_work_item = 0x%x\n",
	     inode->c_work_item));

    // Close it!
    (void) ZwClose (inode->c_handle);
    ObDereferenceObject(inode->c_object);

    // Let them go!
    KeSetEvent (&inode->c_event, EVENT_INCREMENT, FALSE);

    CDEBUG ((D_INODE, "wi_close_handle: exit\n"));
}

// Need to close the c_handle in a system thread 
void inode_close_handle(coda_inode *inode)
{
    NTSTATUS status;
    PDEVICE_OBJECT fsdDevice;
    if (IoIsSystemThread(PsGetCurrentThread())) {
        (void) ZwClose (inode->c_handle);
	ObDereferenceObject(inode->c_object);
    } else {
	// Must call on a worker thread!
	CDEBUG ((D_INODE, "inode_close_handle: Sending out work\n"));
 	inode->c_work_item = IoAllocateWorkItem(coda_device_object);
	ASSERT(inode->c_work_item != NULL);  // Very Bad!  
	IoQueueWorkItem (inode->c_work_item, wi_close_handle,
			 DelayedWorkQueue, (PVOID) inode);
	status = KeWaitForSingleObject (&inode->c_event, Executive,
					KernelMode, FALSE, NULL);
	ASSERT(status == STATUS_SUCCESS);   // Very Bad!

	// Free the work item and clear event
	IoFreeWorkItem (inode->c_work_item);
	KeClearEvent(&inode->c_event);
	CDEBUG ((D_INODE, "inode_close_handle: Work done\n"));
    }

    inode->c_handle = NULL;
    inode->c_object = NULL;
}

//
// Support for the PurgeUid and PurgeFID downcalls
//
void inode_cache_clear_uid(int Uid)
{
    coda_inode *temp = list_entry(inode_all.next, coda_inode, c_alllist);
    int cnt = inode_count;
    int clr_vattr = 0;
    coda_filehandle *tfh;

    while (inode_count-- > 0) {
	if (temp->c_fh_cnt) {
	    tfh = temp->c_fh_list;
	    clr_vattr = FALSE;
	    while (tfh != NULL) {
		if (tfh->fh_uid == Uid) {
		    CDEBUG((D_DOWNCALL,
			    "inode_cache_clear_uid: purge fh %d (%s)\n",
			    tfh->fh_unique, tfh->fh_inode->c_name));
		    if (!NT_SUCCESS(OwPurgeCache(tfh)))
			DbgPrint ("inode_cache_clear_uid: purge error %d\n",
				  tfh->fh_unique);
		    clr_vattr = TRUE;
		}
		tfh = tfh->fh_next;
	    }
	    if (clr_vattr) {
		inode_unsetflag (temp, I_VATTR);
	    }
	}
	temp = list_entry (temp->c_alllist.next, coda_inode, c_alllist);
    }
}

void 
inode_purge_tree (coda_inode *inode, int level)
{
    int cnt;
    coda_inode *child, *tmp;

    CDEBUG ((D_DOWNCALL,
	     "inode_purge_tree: %s (%s) at level %d\n",
	     coda_f2s(&inode->c_fid), inode->c_name, level));

    // postorder processing .. preorder printing :)
    if (inode->c_child_cnt > 0) {
	child = list_entry (inode->c_children.next, coda_inode, c_childlist);
	for (cnt = 0;  cnt < inode->c_child_cnt; cnt++) {
	    tmp = child;
	    child = list_entry (child->c_childlist.next, coda_inode,
				c_childlist);
	    inode_purge_tree (tmp, level+1);
	}
    }

    inode_unsetflag (inode, I_VATTR);
    inode_setflag (inode, I_PURGE);
    inode_cache_delete (inode);
    if ( inode->c_fh_list ) {
	purge_cfh_list (inode);	
    } else {
	inode_release(inode);
    }
}

void inode_replace_fid (coda_inode *inode, CodaFid *oldfid, CodaFid *newfid)
{
    coda_inode *child;
    int ix;

    CDEBUG((D_DOWNCALL, "inode_replace_fid: %s -> %s (name %s)\n",
	    coda_f2s(oldfid), coda_f2s(newfid), inode->c_name));
    inode_cache_delete (inode);
    inode->c_fid = *newfid;
    inode_cache_insert (inode);

    // Recache any direct children.
    if (inode->c_child_cnt > 0) {
	child = list_entry (inode->c_children.next, coda_inode, c_childlist);
	for (ix = 0; ix < inode->c_child_cnt; ix++) {
	    CDEBUG((D_DOWNCALL, "inode_replace_fid: recache child %s (%s)\n",
		    coda_f2s(&child->c_fid), child->c_name));
	    if (!list_empty(&child->c_namelist)) {
		inode_cache_delete (child);
		inode_cache_delete (child);
	    }
	    child = list_entry(child->c_childlist.next, coda_inode,
			       c_childlist);
	}
    }
}


//
// File handle routines 
//

// Get a file handle for the unique dir/inode/uid.
// uid looked up here!
// These are stored in a list in the inode!

coda_filehandle *
get_cfh (coda_inode *inode, coda_inode *dir, int tuid)
{
    coda_filehandle *lfh;

    ENTRY (D_FHAND, "get_cfh");

    // Look for existing one
    lfh = inode->c_fh_list;
    while (lfh != NULL) {
	if (lfh->fh_dir == dir && lfh->fh_uid == tuid) {
	    CDEBUG ((D_FHAND, "get_cfh: Old fh=%d, inode=%s, fh_count=%d, "
		     "fh_uid \n",
		     lfh->fh_unique, coda_f2s(&inode->c_fid),
		     inode->c_fh_cnt, lfh->fh_uid));
	    return lfh;
	}
	lfh = lfh->fh_next;
    }

    if (fh_free == NULL) {
	lfh = (coda_filehandle *) coda_malloc (sizeof(coda_filehandle));
	if (!lfh) {
	    DbgPrint ("get_cfh: out of memory!\n");
	    return NULL;
	}
    } else {
	lfh = fh_free;
	fh_free = fh_free->fh_next;
    }

    // initialize

    lfh->fh_unique = fh_unique++;
    lfh->fh_inode  = inode;
    lfh->fh_dir    = dir;
    lfh->fh_isopen = 0;
    lfh->fh_granted = 0;
    lfh->fh_openflags = 0;
    lfh->fh_lastacc = 0;
    lfh->fh_uid = tuid;
    
    // Inode fields ...
    lfh->fh_next = inode->c_fh_list;
    inode->c_fh_list = lfh;
    inode->c_fh_cnt++;
    inode_pin (inode);

    CDEBUG ((D_FHAND, "get_cfh: New fh=%d, inode=%s, fh_count=%d\n",
	     lfh->fh_unique, coda_f2s(&inode->c_fid), inode->c_fh_cnt));

    return lfh;
}

void
free_cfh (coda_filehandle *ofh)
{
    coda_filehandle *tfh;
    coda_inode      *inode;

    // Unlink from inode list
    inode = ofh->fh_inode;

    CDEBUG ((D_FHAND, "free_cfh: fh=%d, inode=%s, fh_count=%d\n",
	     ofh->fh_unique, coda_f2s(&inode->c_fid), inode->c_fh_cnt));

    if (inode->c_fh_list == ofh) {
	inode->c_fh_list = ofh->fh_next;
    } else {
	tfh = inode->c_fh_list;
	while (tfh->fh_next && tfh->fh_next != ofh)
	    tfh = tfh->fh_next;
	if (tfh && tfh->fh_next) {
	    tfh->fh_next = ofh->fh_next;
	} else {
	    DbgPrint ("free_cfh: BIG PROBLEM IN NEXT LIST!\n");
	}
    }

    // Decrement count
    inode->c_fh_cnt--;

    // Put on free list
    ofh->fh_next = fh_free;
    fh_free = ofh;

    // Unpin the inode if we are last fh.
    if (!inode->c_fh_list)
	inode_unpin(inode);

    EXIT (D_FHAND, "free_cfh");
}


//
// Utility routine to purge all file handles on an inode
//

void
purge_cfh_list (IN coda_inode *inode)
{
    coda_filehandle *tfh, *nfh;
    int numfh;

    // Get the head
    nfh = inode->c_fh_list;
    numfh = inode->c_fh_cnt;

    // Move down the list ...
    while (nfh != NULL && numfh) {
	tfh = nfh;			// Done in this order to make sure
	nfh = nfh->fh_next;		// tfh can be removed from list.
	numfh--;
	CDEBUG((D_FHAND, "purge_cfh_list: flush wrapper cache fh %d (%s)\n",
		tfh->fh_unique, inode->c_name));
	if (!NT_SUCCESS(OwPurgeCache(tfh)))
	    DbgPrint ("purge_cfh_list: can't purge file handle cache for %s\n",
		      inode->c_name);
    }
    if ((nfh != NULL) || (numfh != 0))
	DbgPrint ("purge_cfh_list: c_fh_list and c_fh_list don't agree!\n");
}
    

#if DBG

// Preorder print
void print_inode_tree (coda_inode *dirinode, int level)
{
    int cnt;
    coda_inode *inode;
    char head[100];

    if (level > 98)
	level = 98;

    for (cnt = 0; cnt < level; cnt++)
	head[cnt] = ((cnt % 5) == 4 ? '+' : '.');
    head[level] = 0;
    CDEBUG ((D_PRCACHE, "%s'%s'\n", head, dirinode->c_name));
    if (dirinode->c_child_cnt > 0) {
	inode = list_entry (dirinode->c_children.next, coda_inode,
			    c_childlist);
	for (cnt = 0; cnt < dirinode->c_child_cnt; cnt++) {
	    print_inode_tree (inode, level+1);
	    inode = list_entry (inode->c_childlist.next, coda_inode,
				c_childlist);
	}
    }
}

// Dump Cache!

void print_inode_cache()
{
    struct dllist_head *le, *lh;
    coda_inode *inode;
    int ix;

    CDEBUG ((D_PRCACHE, "print_inode_cache: *****************************\n"));
    CDEBUG ((D_PRCACHE, "  global vattr cntr = %u\n", glo_vattr_cntr));
    CDEBUG ((D_PRCACHE, "  inode count = %d\n", inode_count));
    CDEBUG ((D_PRCACHE, "  free_count = %d\n", free_count));
    if (DebugLevel & D_DETAILS) {
      CDEBUG ((D_PRCACHE, "  Free List:\n"));
      lh = &inode_freelist;
      for ( le = lh->next; le != lh; le = le->next) {
	inode = list_entry(le, coda_inode, c_freelist);
	CDEBUG ((D_PRCACHE, "     %s (%s) fhcount %d,%u,%s%s%s\n",
		 coda_f2s(&inode->c_fid),
		 inode->c_name,
		 inode->c_fh_cnt,
		 inode->c_vattr_cntr,
		 (inode_hasflag(inode, I_REL)? " Rel": ""),
		 (inode_hasflag(inode, I_PINNED)? " Pin": ""),
		 (inode_hasflag(inode, I_VATTR)? " Vattr": "")));
      }
      CDEBUG ((D_PRCACHE, "  FID cache:\n"));
      for (ix = 0; ix < CACHE_SIZE; ix++) {
	if (!list_empty(&(inode_fid_cache[ix]))) {
	  CDEBUG ((D_PRCACHE, "     %d:\n", ix));
	  lh = &(inode_fid_cache[ix]);
	  for ( le = lh->next ; le != lh; le = le->next) {
	    inode = list_entry(le, coda_inode, c_fidlist);
	    CDEBUG ((D_PRCACHE, "     %s (%s) fhcount %d,%u,%s%s%s\n",
		     coda_f2s(&inode->c_fid),
		     inode->c_name,
		     inode->c_fh_cnt,
		     inode->c_vattr_cntr,
		     (inode_hasflag(inode, I_REL)? " Rel": ""),
		     (inode_hasflag(inode, I_PINNED)? " Pin": ""),
		     (inode_hasflag(inode, I_VATTR)? " Vattr": "")));
	  }
	}
      }
      
      CDEBUG ((D_PRCACHE, "  Name cache:\n"));
      for (ix = 0; ix < CACHE_SIZE; ix++) {
	if (!list_empty(&(inode_name_cache[ix]))) {
	  CDEBUG ((D_PRCACHE, "     %d:\n", ix));
	  lh = &(inode_name_cache[ix]);
	  for ( le = lh->next ; le != lh; le = le->next) {
	    inode = list_entry(le, coda_inode, c_namelist);
	    CDEBUG ((D_PRCACHE, "     %s (%s) fhcount %d,%u,%s%s%s\n",
		     coda_f2s(&inode->c_fid),
		     inode->c_name,
		     inode->c_fh_cnt,
		     inode->c_vattr_cntr,
		     (inode_hasflag(inode, I_REL)? " Rel": ""),
		     (inode_hasflag(inode, I_PINNED)? " Pin": ""),
		     (inode_hasflag(inode, I_VATTR)? " Vattr": "")));
	  }
	}
      }
      
      CDEBUG ((D_PRCACHE, "  alllist:\n"));
      lh = &inode_all;
      for ( le = lh->next; le != lh; le = le->next) {
	inode = list_entry(le, coda_inode, c_alllist);
	CDEBUG ((D_PRCACHE, "     %s (%s) fhcount %d,%u,%s%s%s%s\n",
		 coda_f2s(&inode->c_fid),
		 inode->c_name,
		 inode->c_fh_cnt,
		 inode->c_vattr_cntr,
		 (inode_hasflag(inode, I_REL)? " Rel": ""),
		 (inode_hasflag(inode, I_PINNED)? " Pin": ""),
		 (inode_hasflag(inode, I_VATTR)? " Vattr": ""),
		 (inode_hasflag(inode, I_ISCTRL)? " Ctrl": "")));
      }

      if (coda_nt_volume.v_rootinode) {
	  CDEBUG ((D_PRCACHE, "Tree Form:\n"));
	  print_inode_tree (coda_nt_volume.v_rootinode, 0);
      }
    }

    CDEBUG ((D_PRCACHE, "print_inode_cache: END  ************************\n"));
    
}
#endif
